% S = generate_hok_descriptor(X, pivots) generates a HOK descriptor for a
% given feature matrix X and a set of learned pivots.
%
% inputs: X is a nxd matrix with n frames and each of d dimejsional features
% inputs: pivots, represents a pivot set learned for the task. See demo.m for how this is done.
% output: S is a 3D HOK tensor.
% 
% for code or bugs report to anoop.cherian@anu.edu.au.
%
%
% This code is the implementation of the following paper. If you use this 
% code, please cite the paper.
% @inproceedings{hok_pooling,
%     author = {Anoop Cherian and Piotr Koniusz and Stephen Gould},
%     title = {Higher-order Pooling of CNN Features via Kernel Linearization
%     for Action Recognition},
%     booktitle = {Winter Conference on Computer Vision (WACV)},
%     year = {2017}
% }   
% % This code is released on the BSD3 license.
% % This code should not be used for non-commercial purposes. 
% % The authors are not liable to any loss or damage caused by running this code.
% % For any issues/bugs, please contact anoop.cherian@gmail.com.
% 
% Copyright (c) 2017, Anoop Cherian All rights reserved.
%{ 
% Redistribution and use in source and binary forms, with or without modification, 
% are permitted provided that the following conditions are met:
% 1. Redistributions of source code must retain the above copyright notice, 
% this list of conditions and the following disclaimer.
% 2. Redistributions in binary form must reproduce the above copyright notice, 
% this list of conditions and the following disclaimer in the documentation and/or 
% other materials provided with the distribution.
% 3. Neither the name of the copyright holder nor the names of its contributors may 
% be used to endorse or promote products derived from this software without specific 
% prior written permission.
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
% EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
% MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
% COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
% EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
% GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
% AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
% NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
% OF THE POSSIBILITY OF SUCH DAMAGE.
%}
%
% pre-requisites to run this code:
% install liblinear 2.x from https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/multicore-liblinear/
% install tensor toolbox for higher-order SVD from http://au.mathworks.com/matlabcentral/fileexchange/25514-tp-tool
% install vl_feat from http://www.vlfeat.org/
%

function S = generate_hok_descriptor(X, pivots)       
    % temporal pivots
    T = size(X,1);
    z = [0:0.25:1]; sigmaT=0.1; v=(1:T)/T; sigmaX=1e-5; % temporal pivot parameters

    dist1 = mypdist2(z', v');
    mm = pivots.mean./pivots.dcov;
    dist2 = zeros(size(mm,2), size(X,1));
    for t=1:size(mm,2)
        dist2(t,:) = mypdist2(mm(:,t)',X./repmat(pivots.dcov(:,t)',[size(X,1),1]));
    end
    dist = exp(-[sigmaT*dist1; sigmaX*dist2]/2); % spatial and temporal distances to pivots.
 
    % to avoid supersymmetry in the descriptor, find the indicies for HOK that
    % we will use in later.    
    persistent idx;
    n = size(dist,1);
    if isempty(idx)          
         idx = get_idx(n);
    end
    
    % compute HOK tensor via third order outerproduct.
    S0 = vec(dist*dist');
    S = S0*dist(:,1)';
    for i=2:size(dist,2)
        S = S + S0*dist(:,i)';
    end
    S = S/size(dist,2);
    S = reshape(S, [n,n,n]);

    % power normalization
    epn_factor = 0.1;
    [s, U, ~, ~] = hosvd(S, ones(1,3));
    S=tprod1( tprod1( tprod1(sign(s).*(abs(s)).^epn_factor, U{1},1), U{2},2), U{3},3);
    S = vec(S(idx))';
end

% some rudimentary method to find the non-redundant indices of a
% super-symmetric tensor.
function idx = get_idx(n)
   oo = ones(n); oo = repmat(oo, [1,1,n]);
   for i=1:n
       for j=i:n
           for k=j:n
               oo(i,j,k)=1234;
           end
       end
   end
   idx = find(oo==1234);
end

function v=vec(x)
v=x(:);
end
